package week03

import scala.annotation.tailrec

object Practice extends App {

  /**
   * Írjunk egy containsValue függvényt, ami true-t ad vissza, ha az input list: List[Int]-ben
   * szerepel a value: Int és false-t egyébként.
   *
   * pl. ha list = 1.:4::2::8::5::7 és value = 3, akkor az eredmény false.
   */
  @tailrec
  def containsValue(list: List[Int], value: Int): Boolean = list match {
    case Nil => false
    case `value`::_ => true //ez így backtickek közt ugyanaz, mint a head::_ if (head==value) => true
    case _::tail => containsValue(tail, value)
  }

  /**
   * Írjunk egy find függvényt, ami az input list: List[Int]-ben megkeresi az első előfordulását
   * a value: Int-nek és visszaadja a pozíciót! (a head a 0. elem, a rákövetkező az 1. stb).
   * Ha az érték nem szerepel a listában, adjon vissza -1-et.
   *
   * pl. ha list = 1::4::2::4::1::3 és value = 4, akkor az eredmény 1 (mert a 4-es először az 1. pozin szerepel).
   */
  def find(list: List[Int], value: Int): Int = {
    @tailrec
    def find(list: List[Int], pos: Int): Int = list match {
      case Nil => -1
      case `value`::_ => pos
      case _::tail => find(tail,pos+1)
    }
    find(list, 0)
  }

  /**
   * Írjunk egy removeDuplicates függvényt, ami az input list: List[Int]-ből minden érték előfordulásából
   * csak a legelsőt tartja meg!
   *
   * pl. ha list = 1::4::2::4::1::3, akkor az eredmény 1::4::2::3
   */
  def removeDuplicates(list: List[Int]): List[Int] = {
    def pushFirsts(list: List[Int], acc: List[Int]): List[Int] = list match {
      case Nil => acc
      case head :: tail => if (containsValue(acc, head)) pushFirsts(tail, acc) else pushFirsts(tail, head::acc)
    }
    pushFirsts(list, Nil).reverse
  }

  /**
   * Írjunk egy tail recursive függvényt, ami az input n: Intre visszaadja az első n négyzetszám összegét!
   * Ha n <= 0, legyen az eredmény 0.
   */
  def sumSquares(n: Int): Int = {
    @tailrec
    def sumSquares(n: Int, acc: Int): Int = {
      if( n == 0 ) acc else sumSquares(n-1, acc+n*n)
    }
    sumSquares(n, 0)
  }

  /**
   * Írjunk egy függvényt, ami az input list: List[Int]-re visszaadja azt a listát, melyben
   * az eredeti lista minden páratlan elemét négyzetreemeljük, a párosakat békén hagyjuk.
   *
   * pl. ha list = 1::4::2::8::5::7, akkor az eredmény 1::4::2::8::25::49
   */
  def squareOdds(list: List[Int]):List[Int] = list map { x => if (x%2==1) x*x else x }

  /**
   * Írjunk egy függvényt, ami visszaadja, hogy az input n: Int szám prímszám-e!
   * Ha használunk rekurziót, az tail recursion legyen.
   */
  def isPrime(n: Int): Boolean = {
    def divides(m: Int): Boolean = {
      if (m*m > n) false else ((n%m == 0)||divides(m+1))
    }
    if(n<=1) false else !divides(2)
  }

  /**
   * Írjunk egy függvényt, ami az input list: List[Int]-et szétvágja két listára, az egyikbe legyűjti
   * a prímszámokat, a másikba az összetett számokat (és eldobja a negatívakat, a nullát és az egyet)!
   *
   * pl. ha list = 1::4::2::(-5)::0::8::5::7::5, akkor az eredmény ((2::5::7::5),(4::8))
   */
  def splitPrimes(list: List[Int]):(List[Int],List[Int]) = list filter { _>1 } partition isPrime
  //a partition függvény kap egy predikátumot és két listára osztja az eredetit:
  // az elsőbe kerülnek, akikre a predikátum igaz
  // a másodikba, akikre hamis
}
