package week09

trait NamedObject {
  def name: String
}
trait Pet extends NamedObject
case class Cat(name: String) extends Pet
case class Dog(name: String) extends Pet

trait Option[+T] {
  def foreach[U](f: T=>U): Unit
}
case class None[T]() extends Option[T] {
  override def foreach[U](f: T => U): Unit = ()
}
case class Some[T](value: T) extends Option[T] {
  override def foreach[U](f: T => U): Unit = f(value)
}

case class Ember(name: String, catOption: Option[Cat], dogOption: Option[Dog]) extends NamedObject

object Main extends App {
  def printName(named: NamedObject) = println(named.name)
  def printNameOption(namedOption: Option[NamedObject], prefix: String) =
    for(named <- namedOption) {
      print(prefix)
      printName(named)
    }

  def printEmber(ember: Ember) = {
    printName(ember)
    printNameOption(ember.catOption, "Macskája: ")
    printNameOption(ember.dogOption, "Kutyája: ")
  }

  val morcos = Cat("Morcos")
  val freya  = Dog("Freya")
  val tibiNeni = Ember("Tibi", Some(morcos), None() )

  printEmber(tibiNeni)

}