package week09

import scalafx.Includes._
import scalafx.application.JFXApp
import scalafx.application.JFXApp.PrimaryStage
import scalafx.beans.property.DoubleProperty.sfxDoubleProperty2jfx
import scalafx.beans.property.ObjectProperty
import scalafx.scene.canvas.Canvas
import scalafx.scene.input.{KeyCode, KeyEvent, MouseButton, MouseEvent, ScrollEvent}
import scalafx.scene.paint.Stop.sfxStop2jfx
import scalafx.scene.paint.{Color, CycleMethod, LinearGradient, Paint, Stop}
import scalafx.scene.shape.Rectangle
import scalafx.scene.{Group, Scene}

class Model {
  val actorVector = ObjectProperty(this, "actorVector", ActorVector(Port(Point(0,0),100,100), Vector()) )
}

object HelloPaint extends JFXApp {
  val canvas = new Canvas(360, 360)
  val model = new Model()

  // Draw background with gradient
  val rect = new Rectangle {
    height = 400
    width = 400
    fill = new LinearGradient(0, 0, 1, 1, true, CycleMethod.Reflect, List(Stop(0, Color.Red), Stop(1, Color.Yellow)))
  }

  val rootPane = new Group
  rootPane.children = List(rect, canvas)

  stage = new PrimaryStage {
    title = "Objektum kattintós GUIs scalafx app"
    scene = new Scene(400, 400) {
      root = rootPane
      onKeyPressed = (e: KeyEvent) => {
        e.code match {
          case KeyCode.Left => model.actorVector.value = model.actorVector.value.translate(10,0)
          case KeyCode.Right => model.actorVector.value = model.actorVector.value.translate(-10,0)
          case KeyCode.Up => model.actorVector.value = model.actorVector.value.translate(0,10)
          case KeyCode.Down => model.actorVector.value = model.actorVector.value.translate(0,-10)
          case _ => ()
        }
        reset
      }
    }
  }

  canvas.translateX = 20
  canvas.translateY = 20
  val gc = canvas.graphicsContext2D
  reset

  canvas.onMouseClicked = (e: MouseEvent) => {
    if( e.isControlDown ){
      model.actorVector.value = model.actorVector.value.removeShapesAt(e.x - canvas.width.get/2,e.y-canvas.width.get/2)
    }else if( e.isShiftDown ){
      model.actorVector.value = model.actorVector.value.addRandomShape
    }else{
      model.actorVector.value = model.actorVector.value.flipShapesAt(e.x - canvas.width.get/2,e.y-canvas.height.get/2)
    }
    reset
  }
  canvas.onScroll = (e: ScrollEvent) => {
    val ratio = if(e.deltaY < 0) 0.9 else 1.1
    model.actorVector.value = model.actorVector.value.zoom(ratio)
    reset
  }

  private def draw(coloredShape: ColoredShape): Unit = coloredShape.shape match {
    case Kor(x,y,r) => {
      gc.stroke = coloredShape.draw
      gc.lineWidth = 5
      gc.fill = coloredShape.fill
      gc.strokeOval(canvas.width.get/2 + x - r, canvas.height.get/2 + y - r, 2 * r, 2 * r)
      gc.fillOval(canvas.width.get/2 + x - r, canvas.height.get/2 + y - r, 2 * r, 2 * r)
    }
    case Teglalap(x,y,width,height) => {
      gc.stroke = coloredShape.draw
      gc.lineWidth = 5
      gc.fill = coloredShape.fill
      gc.strokeRect(canvas.width.get/2 + x, canvas.height.get/2 + y,width,height)
      gc.fillRect(canvas.width.get/2 + x, canvas.height.get/2 + y,width,height)
    }
  }

  private def reset(): Unit = {
    gc.fill = Color.Black
    gc.globalAlpha = 1.0
    gc.fillRect(0, 0, canvas.width.get, canvas.height.get)
    val offX = canvas.width.get/2
    val offY = canvas.height.get/2

    for( coloredShape <- model.actorVector.value.shapes ) {
      draw(coloredShape)
    }

    val port = model.actorVector.value.canvas
    val upperLeft = port.center + Vektor(-port.width,-port.height)
    gc.fill = Color.Yellow
    gc.globalAlpha = 0.1
    gc.fillRect( upperLeft.x + offX, upperLeft.y + offY, port.width * 2, port.height * 2)
  }

}
