package week11

import java.text.NumberFormat

import scala.collection.parallel.CollectionConverters._

object ParMain extends App {
  def time[U](n: Int, message: String)(block: =>U): U = {
    def sumTime(n: Int, acc: Long, value: => U): (Long,U) = n match {
      case 0 => (acc,value)
      case _ => {
        val t0 = System.nanoTime()
        val result = block
        val t1 = System.nanoTime()
        sumTime(n-1, acc+(t1-t0),result)
      }
    }
    val (totalTime,result) = sumTime(n,0,block)
    val averageTime = totalTime / n
    println("Elapsed time (" +message+ "): " + NumberFormat.getInstance().format(averageTime) + " ns, averaged over " + n + " calls")
    result
  }

  val seqVector : Vector[Int] = (1 to 100).toVector
  time(10,"convert to par") (seqVector.par)
  val parVector = seqVector.par

  def f: Int => Int =  x => {Thread.sleep(10);x+1}
  def p: Int => Boolean = x => {Thread.sleep(10); x % 3 == 0}
  def op: (Int,Int) => Int = (x,y) =>{Thread.sleep(10); y+x+y%3 }

  time(10, "sequential map") (seqVector map f)
  time(10, "parallel map") (parVector map f)
  val parMapped = parVector map f
  time(10, "convert to seq") (parMapped.seq )

  time(10, "sequential filter") (seqVector filter p)
  time(10, "parallel filter") (parVector filter p)
  val parFiltered = parVector filter p
  time(10, "convert to seq") (parFiltered.seq)

  println (time(10, "sequential reduce")(seqVector reduce op))
  println (time(10, "sequential reduceLeft")(seqVector reduceLeft  op))
  println (time(10, "sequential reduceRight")(seqVector reduceRight op))
  println( time(10, "parallel reduce")(parVector reduce op))
  println( time(10, "parallel reduceLeft")(parVector reduceLeft op))
  println( time(10, "parallel reduceRight")(parVector reduceRight op))
}
