package week11.stream

import java.io.PrintWriter

import scala.io.Source
import scala.util.{Failure, Success, Try}

/**
 * Feladat: implementáljunk
 * - Ember osztályt: String ID, String vezeteknev, String keresztnev, String nick, int szint
 * - Műkincs osztályt: String ID, String név, Int méret, Int érték
 * - Láda osztályt: String ládaID, String emberID, Int méret, Vector[Mukincs] tartalom
 *
 * - az Ember osztályba egy save(PrintWriter): Unit metódust, ami egy megnyitott file-ba kiírja ezeket *egy sorba, enter nélkül*, (mondjuk) pontosvesszővel elválasztva!
 *   (kezeljük le valahogy azt az esetet is, amikor is van pontosvessző valamelyik String mezőben)
 * - az Ember companion objectbe egy load(String):Try[Ember] metódust, ami egy, a save metódussal kimentett Ember objektumot betölt és visszaad Successként,
 *   ha a String egy embert kódol és egy  InvalidClassException -t valamennyi infóval a hibáról, ha nem!
 *
 * - Műkincs osztályra ugyanezeket
 *
 * - Lada osztályra is ugyanezeket azzal, hogy a Vector[Mukincs]-et a Mukincs objektumok egymás után írásával ;-kkel oldjuk meg!
 *
 * - egy metódust, ami beolvas egy filet a következő kódolásban:
 *   Emberek //ez egy string
 *   Ember1
 *   Ember2
 *   ...
 *   Embern
 *   Ladak //ez egy String
 *   Lada1
 *   Lada2
 *   ...
 *   Ladam
 *
 * - üres sorokat tegyük közben skipre
 * - kivételeket, amik közben jönnek láda- vagy emberolvasáskor, gyűjtsük egy listába
 * - tegyünk bele új kivételosztályt is, ami a DuplicateID névre hallgat, eltárolja a GameObject-et és valami meaningful string üzenetet ad vissza!
 * - adjunk vissza: egy Map[String,(Ember,Vector[Lada])] mapet az emberekkel, emberID-hez rendelve, mindenkihez kigyűjtve, hogy melyek az ő ládái,
 * továbbá egy Vector[Lada]-t a "gazdátlan" ládákkal, melyekhez tartoz emberID nem létezik, továbbá a kivételek listájával!
 *
 * - Írjunk LadaOptimizer objektumot, mely egy Vector[Lada]-ból egy Vector[Lada]-t készít úgy, hogy az összes műkincset megtartva megpróbálja
 *   minimalizálni a nemüres ládák számát! Heurisztikus megoldás is jó, garantáltan optimális megoldás is jó.
 *
 * - Írjunk egy olyan mainLoop-ot, ami
 *   - bekér a felhasználótól egy input- és egy output filenevet
 *   - ha sikerül beolvasnia az inputból egy fenti formátumú file-t, akkor (bezárja, majd) egy külön szálon futtatva kioptimalizálja a minden emberhez
 *     tartozó ládák tartalmát, kiürítve a lehető legtőbbet, majd az eredményt visszaírja a kimenetként kapott file-ba! Siker esetén
 *     írja ki a kiürített ládák számát és hogy mentve (ha sikerült), valamint az esetleg az input feldolgozása vagy a file mentése közben jövő kivételeket!
 *     A beolvasást is külön szálon futtassa és jöjjön vissza a konzolra egy eredménnyel, ekkor írva ki, hogy sikerült-e beolvasni vagy sem az input file
 *     minden egyes sorát.
 *   - amíg a számítás tart, a felhasználó hadd adjon meg másik file-t is.
 *
 *
 * Ettől egy független feladatként: implementáljunk lazy listával egy generatePrimes metódust, ami egy végtelen, lazy listát generál,
 * melyen pl a for( n <- generatePrimes.take(1000)) println(n) hívással kiírjuk az első 1000 prímszámot!
 *
 */

case class Ember(id: String, vezeteknev: String, keresztnev: String, nev: String, szint: Int){
  def save(writer: PrintWriter): Unit = {
    writer.println( s"$id;$vezeteknev;$keresztnev;$nev;$szint" )
  }
}
object Ember {
  def load(line: String): Try[Ember] = {
    val fields: Array[String] = line.split(";") // String[] fields = line.split(";")
    //verify: 5 mező van-e nem => exception; igen => szint-ből intet készítünk
    //ha mind sikerül => Success(Ember(...)) ha nem => Failure(exception)
    if(fields.length == 5) {
      Try(Ember(fields(0),fields(1),fields(2),fields(3),fields(4).toInt))
        .transform( ember => Success(ember) ,
          {
            case nfe => Failure(new IllegalArgumentException("Nem integer: " + fields(4)))
            case e => Failure(e)
          }
        )
    } else Failure(new IllegalArgumentException( s"Ember: mezők száma nem megfelelő: $line" ))
  }
}

case class Mukincs(id: String, nev: String, meret: Int, ertek: Int) {
  def save(printWriter: PrintWriter): Unit =
    printWriter.print(s"$id;$nev;$meret;$ertek")
}
object Mukincs {
  def load(first: String, second: String, third: String, fourth: String ): Try[Mukincs] = {
    Try(Mukincs(first,second,third.toInt,fourth.toInt))
  }
}

case class Lada(id: String, emberId: String, meret: Int, tartalom: Vector[Mukincs]) {
  def save(writer: PrintWriter): Unit = {
    writer.print(s"$id;$emberId;$meret;")
    for( mukincs <- tartalom ) {
      mukincs.save(writer)
      writer.print(";")
    }
  }
}

object Lada {
  def load(line: String): Try[Lada] = {
    val fields = line.split(";")
    if(fields.length % 4 == 3) { // de hogyan oldjuk meg azt, hogy négyesével jó indexekről kapja az értékeket?
      val mukincsVektor = (3 until fields.length by 4).map(
        i => Mukincs.load(fields(i), fields(i + 1), fields(i + 2), fields(i + 3))
      ).toVector.flatMap(_.toOption)
      // Vector[Try[Mukincs]] ==> Vector[Mukincs] uh ami Success(mk), abból legyen mk
      Try(Lada(fields(0), fields(1), fields(2).toInt, mukincsVektor))
    }else Failure(new IllegalArgumentException(s"Lada: Nem megfelelő mezőszám: $line"))
  }
}

object EmberLadaLoader {
  def process(source: Source): Map[String, (Ember, Vector[Lada])] = {
    //vissza kell adnom valamit, és minden sorral amit csinálunk, az függhet az előzőektől => fold
    // ("" / "Emberek" / "Ladak" , Vector[Ember], Vector[Ladak] )
    val (_,emberVektor,ladaVektor) =
      source.getLines().foldLeft[(String, Vector[Ember],Vector[Lada])]( ("",Vector(), Vector()) ){
      (acc , line) => line match {
        case "Emberek" => acc.copy( _1 = "Emberek") //("Emberek", acc._2, acc._3 )
        case "Ladak" => acc.copy( _1 = "Ladak")
        case _ => acc._1 match {
          case "Emberek" =>
            acc.copy( _2 = acc._2 :++ Ember.load(line).toOption )
          case "Ladak" =>
            acc.copy( _3 = acc._3 :++ Lada.load(line).toOption )
          case _ => acc
        }
      }
    }
    //ebből a tripletből kéne egy Map-et készíteni, pl úgy, hogy az összes Ember benne van Map[String,Ember]
    val emberMap = emberVektor.foldLeft[Map[String,(Ember,Vector[Lada])]]( Map() ) {
      (acc, ember) => acc.updated(ember.id, (ember,Vector()))
    }
    ladaVektor.foldLeft[Map[String,(Ember, Vector[Lada])]]( emberMap ) {
      (acc,lada) => if(acc.contains(lada.emberId)) {
        //beszúrjuk
        val oldPair = acc(lada.emberId)
        acc.updated(lada.emberId,
          oldPair.copy( _2 = oldPair._2 :+ lada )
        )
      }else acc
    }
  }
  def load(path: String): Map[String, (Ember, Vector[Lada])] = {
    val sourceTry = Try(Source.fromFile(path))
    val ret = (for( source <- sourceTry ) yield process(source))
    for(source <- sourceTry ) {
      source.close()
    }
    ret.getOrElse( Map() )
  }
}

object Gyakorlat extends App {
  val mk = Mukincs("mkID1", "mona lisa", 4, 100)
  val mk2 = Mukincs("mkID1", "fake mona lisa", 4, 1)
  val mk3 = mk.copy(nev = "jobb fake mona lisa", ertek = 5)
  val lada = Lada("ladaID", "valakiID", 10, Vector(mk,mk2,mk3))
  val ember = Ember("valakiID", "James", "Bond", "tibi", 14)
  val ember2 = Ember("valakiMas", "John", "Bond", "tibike", 14)

  val path = "src/main/scala/week11/emberLadaTest.txt"
  val writer = new PrintWriter(path)
  writer.println("Emberek")
  ember.save(writer)
  ember2.save(writer)
  writer.println("Ladak")
  lada.save(writer)
  writer.close()

  val readMap = EmberLadaLoader.load(path)
  for( entry <- readMap ){
    println("Ember:")
    println(entry._2._1)
    println("Ládái:")
    for( lada <- entry._2._2 ) println(lada)
  }
  //println(readMap)
}
