package week12

import java.io.PrintWriter

import scala.io.Source
import scala.util.Try

/**
 * Gyakorlás: a Composite trait objektumok "hierarchikus" struktúrában tárolását
 * (mint pl. Országban Megye, abban Város, abban Kerület, abban Utca, abban Ház)
 * támogatja.
 *
 * A feladat készíteni egy, a fentinek megfelelő Ország..Ház kompozit csomagot,
 * benne már lekérdező és (funkcionálisan) módosító művelettel:
 *
 */

/**
 * A Composite trait, ezt extendeli minden osztály a hierarchiában.
 * A ???-ek csak azért vannak benne most, hogy forduljon a feladatleírás, implementáláskor
 * egyes metódusokat az Orszag, Varos stb. case classokban lehet érdemes implementálni, de lehet, hogy nem mindet
 */
trait Composite {
  def children(): Iterable[Composite] = ??? //adja vissza a csomópont gyerekeit valamiféle iterabléban
  def isLeaf: Boolean = ??? //adja vissza, hogy vannak-e gyerekei a csomópontnak
  def foreach[U](f: Composite => U): Unit = ??? //hajtsa végre a csomóponton és összes leszármazottján az f függvényt
  def map(f: Composite => Composite): Composite = ??? //minden csomóponton hajtsa végre az f függvényt a hierarchiában,
  //a csomópont gyerekei legyenek: az f szerint akik lennének, PLUSZ az eredeti gyerekek képei!
}

/**
 * Az Orszag case class: van neki neve, fővárosa, alatta pedig városok.
 * Konstruktorban ellenőrizzük, hogy
 * - az ország neve nem üres és
 * - a capital benne van valamelyik megyében mint város és
 * - a megyék nevei páronként különböznek!
 * Ha bármelyik feltétel nem teljesül, dobjunk IllegalArgumentException-t.
 */
case class Orszag(name: String, capital: Varos, megyek: Iterable[Megye]) extends Composite {

}

/**
 * Implementáljuk a Megye, Varos, Kerulet, Utca, Haz case classokat is!
 */
/**
 * A Megye: legyen neki String neve, String roviditese, Varos szekhelye és Iterable[Varos] városai!
 * Konstruktorban ellenőrizzük, hogy
 * - a neve nemüres,
 * - a városok nevei páronként különböznek, és
 * - a székhely szerepel a városlistában.
 * Ha bármelyik feltétel nem teljesül, dobjunk IllegalArgumentException-t.
 */
case class Megye()

/**
 * A Varos: legyen neki String neve és Iterable[Kerulet] kerületei!
 * Konstruktorban ellenőrizzük, hogy
 * - a neve nemüres,
 * - kaptunk legalább egy kerületet
 * - a különböző kerületekben az utcák nevei közt nincs azonos
 * - a kerületek számai páronként különböznek és 1-től indulnak egyesével (de magában az Iterabléban nem baj, ha nem
 * "sorrendben" vannak, hiszen lehet, hogy pl egy HashSet-et kapunk)
 * Ha bármelyik feltétel nem teljesül, dobjunk IllegalArgumentExceptiont.
 */
case class Varos()

/**
 * A Kerulet: legyen neki szam: Int sorszáma, String neve és Iterable[Utca] utcái!
 * Konstruktorban ellenőrizzük, hogy
 * - a sorszám pozitív
 * - a név nemüres string
 * - kaptunk legalább egy utcát
 * - és az utcák neve páronként különböző!
 * Ha bármelyik feltétel nem teljesül, dobjunk IllegalArgumentExceptiont.
 */
case class Kerulet()

/**
 * Az Utca: legyen neki nev: String neve és Iterable[Haz] házai!
 * Konstruktorban ellenőrizzük, hogy a név nemüres és hogy a házak helyrajzi számai páronként különböznek.
 */
case class Utca() {

}

/**
 * A Haz: legyen neki Int házszáma, Int helyrajzi száma, Int lakosok száma, String színe!
 * Konstruktorban ellenőrizzük, hogy a lakosok száma nemnegatív,
 * ha negatívat kapnánk, akkor 0 lakossal példányosítsuk a házat!
 * Ellenőrizzük még, hogy a házszám, helyrajzi szám pozitívak, és a szín nemüres string, ha ezek közül
 * bármelyik nem teljesül, dobjunk IllegalArgumentException-t.
 */
case class Haz()


/**
 * A Census objektum egy országra vonatkozó lekérdezéseket futtat:
 */
object Census {
  def countLakosok(regio: Composite): Int = ??? // adja vissza a regio-ban lakó összes embert! pl egy kerületben a benne lévő utcákban lévő házakban lakók számát kell visszaadja.
  def collectSzinek(regio: Composite): Set[String] = ??? //gyűjtse össze a régióban lévő házak színeit egy halmazba!
  def collectUresUtcak(regio: Composite): Set[Utca] = ??? //gyűjtse össze a régióban lévő olyan utcákat, melyekben nincs lakos! (tehát ha van is benne ház, az mind üres)
  def countHazakByColors(regio: Composite): Map[String,Int] = ??? //gyűjtse össze, hogy a régióban lévő házszínek melyikéből mennyi van! (pl sárga házból 2, kékből 16, like that)
  def listSzekhelyek(orszag: Orszag): Vector[Varos] = ??? //adja vissza az ország megyéinek székhelyeit, név szerint ábécébe rendezve!
  def elsoHianyzoHelyrajziSzam(varos: Varos): Int = ??? //adja vissza azt az első pozitív helyrajzi számot, ami a városban még nincs kiosztva háznak
  def getDuplicateHelyrajziszam(varos: Varos): Option[Int] = ??? //adjon vissza (Some-ba wrappelve) egy olyan helyrajzi számot, ami kétszer előfordul a városban vagy None-t, ha nincs ilyen
}

/**
 * Saver, Loader: file IO
 *
 * A Saver kap egy PrintWritert, ebbe ment ki egy országot. Soronként írjuk ki az egyes Composite-okat, annyi .-tal bevezetve, mint amilyen mélyen van
 * a hierarchiában (tehát az Orszag nem kezdődik ponttal, a Megye egy ponttal, a Varos két ponttal stb.)
 * Az egyes sorokba ;-vel elválasztva írjuk ki a primitív adattagokat, majd a következő sorba beljebb tabulálva a nem-primitíveket ha vannak,
 * majd végül a gyerekeket! A megyéket úgy írjuk ki, hogy városaik közül a székhely kerüljön előre, a megyeszékhelyet ne írjuk ki külön!
 *
 * Pl. ha az országnak Budapest a fővárosa, Csongrád egy megyéje, benne két város Szeged (a székhely) és Ópusztaszer,sstb, akkor a
 * file-ba kiírás eredménye ilyesmi lehet:
 *
 * Magyarország
 * .Budapest
 * ..1;Első kerület
 * ...Viktor sugárút
 * ....1;1;4;Arany
 * .Csongrád;CSO
 * ..Szeged
 * ...1;Felsőváros
 * ....Kovács utca
 * .....1;537;4;Barna
 * .....2;514;2;Fehér
 * ....Szabó utca
 * .....1;64;1;Fehér
 * .....3;15;2;Zöld
 * ...2;Alsóváros
 * ....Pék utca
 * .....4;111;0;Tégla
 * ..Ópusztaszer
 * ...1;Ópusztaszer belterület
 * ....Kovács utca
 * .....1;14;2;Barna
 */
case class CensusSaver(writer: PrintWriter) {
  def save(orszag: Orszag, depth: Int): Try[Unit] = ??? //mentsük ki az egész országot a fent megadott formátumban, de csak a depth-ben megadott mélységig!
  //tehát ha pl depth == 0, akkor csak az ország adatai kerüljenek kiírásra, ha depth == 2, akkor az ország, megyék, városok.
}

case class CensusLoader(source: Source) {
  def load(): Try[Composite] = ??? //próbáljunk meg betölteni egy Composite-ot (azzal a feltételezéssel, hogy a pontok száma megadja a típusát)! Ha sikerül ,adjuk vissza egy Success-ban, ha bármilyen formátum probléma jön, Failure
}

object Practice {

}
