package week10

import org.scalatest.concurrent.{Signaler, TimeLimitedTests}
import org.scalatest.exceptions.TestFailedException
import org.scalatest.flatspec.AnyFlatSpec
import org.scalatest.time.{Millis, Span}

import scala.collection.immutable.HashMap

class ZHTest  extends AnyFlatSpec with TimeLimitedTests {
  val timeLimit: Span = Span(1000, Millis)
  override val defaultTestSignaler: Signaler = new Signaler {
    override def apply(testThread: Thread): Unit = {
      println("Ez a teszt túl sokáig fut.")
      testThread.stop()
    } //unsafe, never használd.
  }

  "Rational constructor" should "throw an exception if and only if nevezo is not positive" in {
    for( n <- -10 to 10; m <- -10 to 10 ) {
      try {
        val r = Rational(n, m)
        assert(m >= 0, "No exception thrown for szamlalo = " + n + ", nevezo = " + m)
      }catch{
        case e : TestFailedException => throw e;
        case _ : IllegalArgumentException => assert(m <= 0)
        case e => assert(false, "Wrong exception thrown: " + e )
      }
    }
  }

  "Rational plus" should "add two rational numbers" in {
    for(n1 <- 0 to 5; m1 <- 1 to 5; n2 <- 0 to 5; m2 <- 1 to 5 ){
      val r = Rational(n1,m1) + Rational(n2,m2)
      assert( r.szamlalo*m1*m2 == (n1*m2+n2*m1)*r.nevezo, "teszt: " + n1 + "/" + m1 + " + " + n2 + "/"+ m2)
    }
  }
  "Rational minus" should "subtract two rational numbers" in {
    for(n1 <- 0 to 5; m1 <- 1 to 5; n2 <- 0 to 5; m2 <- 1 to 5 ){
      val r = Rational(n1,m1) - Rational(n2,m2)
      assert( r.szamlalo*m1*m2 == (n1*m2-n2*m1)*r.nevezo, "teszt: " + n1 + "/" + m1 + " - " + n2 + "/"+ m2)
    }
  }
  "Rational product" should "multiply two rational numbers" in {
    for(n1 <- 0 to 5; m1 <- 1 to 5; n2 <- 0 to 5; m2 <- 1 to 5 ){
      val r = Rational(n1,m1) * Rational(n2,m2)
      assert( r.szamlalo*m1*m2 == n1*n2*r.nevezo, "teszt: " + n1 + "/" + m1 + " + " + n2 + "*"+ m2)
    }
  }
  "Rational division" should "throw an IllegalArgumentException if and only if the divisor is zero " in {
    for(n1 <- 0 to 5; m1 <- 1 to 5; n2 <- 0 to 5; m2 <- 1 to 5 ){
      try {
        val r = Rational(n1, m1) / Rational(n2, m2)
        assert(n2 > 0)
      }catch {
        case _ : IllegalArgumentException => assert(n2 == 0)
        case e => throw e
      }
    }
  }
  "Rational division" should "compute the result of the division, duh " in {
    for(n1 <- -5 to 5; m1 <- 1 to 5; n2 <- -5 to 5 if n2 != 0; m2 <- 1 to 5 ){
      val r = Rational(n1,m1) / Rational(n2,m2)
      assert( r.nevezo * n1 * m2 == m1 * n2 * r.szamlalo , "teszt: " + n1 + "/" + m1 + " + " + n2 + "*"+ m2)
    }

  }
  "Rational unary minus" should "negate its argument" in {
    for(n <- -5 to 5; m <- 1 to 5) {
      val r = - Rational(n,m)
      assert( r.szamlalo * m == -n * r.nevezo, "test: " + n + "/" + m )
    }
  }
  "Rational comparison" should "compare two rationals" in {
    for(n1 <- 0 to 5; m1 <- 1 to 5; n2 <- 0 to 5; m2 <- 1 to 5 ){
      val r = Rational(n1,m1) compare Rational(n2,m2)
      val sz = n1 * m2 - n2 * m1
      assert(
        sz * r > 0 || (sz == 0 && r == 0),
        "teszt: " + n1 + "/" + m1 + " + " + n2 + "*"+ m2)
    }
  }
  "Rational equals" should "evaluate to true if the rationals are the same" in  {
    for(n1 <- 0 to 5; m1 <- 1 to 5; n2 <- 0 to 5; m2 <- 1 to 5 ) {
      assert(
        (Rational(n1,m1)==Rational(n2,m2))
        ==(n1*m2 == n2*m1),
        "teszt: " + n1 + "/" + m1 + " + " + n2 + "*"+ m2
      )
    }
  }
  "Rational equals" should "evaluate to false when compared to a non-Rational object" in {
    assert(!(Rational(1, 1) == "sanyi"))
    assert(!(Rational(1, 1) == null))
    assert(!(Rational(1, 1) == 4))
  }

  "Rational toString" should "return 'szamlalo'/'nevezo'" in {
    for(n1 <- 0 to 5; m1 <- 1 to 5){
      assert( Rational(n1,m1).toString == n1 + "/" + m1)
    }
  }

  "ValtozoErtekadas()" should "return an empty ValtozoErtekadas" in {
    assert( ValtozoErtekadas().map == HashMap[String,Rational]())
  }

  "ValtozoErtekadas().get(x)" should "return None" in {
    for( c <- "xyzuv") {
      assert( ValtozoErtekadas().get(c.toString).isEmpty)
    }
  }

  "ValtozoErtekadas().set(x,r).get(x)" should "return Some(r) " in {
    for( c <- "xyzuv"; n1 <- -5 to 5 ) {
      assert(
        ValtozoErtekadas()
          .set(c.toString, Rational(n1, 1))
          .get(c.toString)
          .contains(Rational(n1, 1)),
        "teszt: " + c + ", r = " + n1
      )
    }
  }

  "ValtozoErtekadas().set(x,r).unset(x).get(x)" should "return None" in {
    for( c <- "xyzuv"; n1 <- -5 to 5 ) {
      assert(
        ValtozoErtekadas()
          .set(c.toString, Rational(n1, 1))
          .unset(c.toString)
          .get(c.toString)
          .isEmpty,
        "teszt: " + c + ", r = " + n1
      )
    }
  }

  "ValtozoErtekadas().set(x,r).set(y,r).isSet(z)" should "return Some(r)" +
    "if z is x or y and None otherwise" in {
    for (c <- "xyzuv"; d <- "xyzuv"; n1 <- -5 to 5; e <- "xyzuv") {
      assert(
        ValtozoErtekadas()
          .set(c.toString, Rational(n1, 1))
          .set(d.toString, Rational(n1, 1))
          .isSet(e.toString)
          == (e == c || e == d),
        "teszt: " + c + ", " + d + ", " + n1 + " isSet: " + e
      )
    }
  }

  "ValtozoErtekadas().map f" should "evaluate to ValtozoErtekadas()" in {
    assert( ValtozoErtekadas().map { r => Rational(r.nevezo,r.szamlalo) } ==
      ValtozoErtekadas() )
  }

 "ValtozoErtekadas().set(x,r1).set(y,r2).map( x => 2 * x )" should "assign x->2r1 and y->2r2" in {
   for( c <- "xyzuv"; d <- "xyzuv"; n1 <- -5 to 5 ) {
     val res =
       ValtozoErtekadas()
         .set(c.toString, Rational(n1, 1))
         .set(d.toString, Rational(n1, 1))
         .map{ r => Rational(2*r.szamlalo , r.nevezo)}
     assert( res.get(c.toString) contains Rational (2*n1,1) )
     assert( res.get(d.toString) contains Rational (2*n1,1) )
     assert( res.get("a").isEmpty )
   }
  }
  "ValtozoErtekadas().foreach( println( _._1) )" should "not print anything" in {
      val s = new StringBuilder()
        ValtozoErtekadas()
          .foreach( entry => (s append entry._1) )
      assert( s.toString == "")
  }

  "ValtozoErtekadas().set(x,r1).set(y,r2).foreach( println( _._1) )" should "print the two keys" in {
    for( c <- "xyzuv"; d <- "xyzuv"; n1 <- -5 to 5 ) {
      val s = new StringBuilder()
        ValtozoErtekadas()
          .set(c.toString, Rational(n1, 1))
          .set(d.toString, Rational(n1, 1))
          .foreach( entry => (s append entry._1) )
      assert( s.toString contains c )
      assert( s.toString contains d )
      assert( !(s.toString contains 'w'))
    }
  }

  "A rational" should "evaluate to itself" in {
    assert( Rational(1,1)(ValtozoErtekadas()) contains Rational(1,1))
  }

  "Adding two rationals in an expression" should "produce their sum" in {
    assert(
      Add(Rational(1, 2), Rational(1, 3))(ValtozoErtekadas())
        .contains(Rational(5, 6))
    )
  }
  "Subtracting two rationals in an expression" should "produce their difference" in {
    assert(
      Sub(Rational(1, 2), Rational(1, 3))(ValtozoErtekadas())
        .contains(Rational(1, 6))
    )
  }
  "Multiplying two rationals in an expression" should "produce their product" in {
    assert(
      Mul(Rational(1, 2), Rational(1, 3))(ValtozoErtekadas())
        .contains(Rational(1, 6))
    )
  }
  "Dividing two rationals in an expression" should "produce their quotient" in {
    assert(
      Div(Rational(1, 2), Rational(1, 3))(ValtozoErtekadas())
        .contains(Rational(3, 2))
    )
  }
  "Dividing by zero in an expression" should "produce None" in {
    assert(
      Div(Rational(1, 2), Rational(0, 3))(ValtozoErtekadas())
        .isEmpty
    )
  }
  "Adding x and y when both are one" should "produce two" in {
    assert(
      Add(Variable("x"), Variable("y"))(
        ValtozoErtekadas().set("x",Rational(1,1)).set("y",Rational(1,1))
      ) contains Rational(2,1)
    )
  }
  "Adding x and y when x=1 and y is unset" should "produce None" in {
    assert(
      Add(Variable("x"), Variable("y"))(
        ValtozoErtekadas().set("x",Rational(1,1))
      ).isEmpty
    )
  }
  "Evaluating x/y when x = 1 and y = 0 " should "produce None" in {
    assert(
      Div(Variable("x"), Variable("y"))(
        ValtozoErtekadas().set("x",Rational(1,1)).set("y",Rational(0,1))
      ).isEmpty
    )
  }
  "Evaluating a complicated expression" should "produce the correct result" in {
    assert(
      Mul(Variable("x"), Add(Variable("y"), Rational(1,2)))(
        ValtozoErtekadas().set("x",Rational(1,1)).set("y",Rational(2,1))
      ) contains Rational(5,2)
    )
  }
}
