package week13

import java.io.PrintWriter

import org.scalatest.concurrent.{Signaler, TimeLimitedTests}
import org.scalatest.flatspec.AnyFlatSpec
import org.scalatest.time.{Millis, Span}

import scala.io.Source
import scala.util.Try

class ZHTest extends AnyFlatSpec with TimeLimitedTests {
    val timeLimit: Span = Span(1000, Millis)
    override val defaultTestSignaler: Signaler = new Signaler {
      override def apply(testThread: Thread): Unit = {
        println("Ez a teszt túl sokáig fut.")
        testThread.stop()
      } //unsafe, never használd.
    }

    "Doboz konstruktor" should "work if there are no problems" in {
      val doboz = Doboz(10,8,"Piros",Vector())
      //this should work.
    }

    it should "throw an IllegalArgumentException if szin is an empty string" in {
      val tryConstruct = Try(Doboz(10,8,"",Vector()))
      assert(tryConstruct.isFailure)
      val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
      assert(tryRecover.isSuccess)
    }
  it should "throw an IllegalArgumentException if szin is an empty string - 2" in {
    val tryConstruct = Try(Doboz(10,8,"",Vector()))
    assert(tryConstruct.isFailure)
    val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
    assert(tryRecover.isSuccess)
  }

  it should "throw an IllegalArgumentException if kivul < 1" in {
    val tryConstruct = Try(Doboz(0,0,"Piros",Vector()))
    assert(tryConstruct.isFailure)
    val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
    assert(tryRecover.isSuccess)
  }
  it should "throw an IllegalArgumentException if kivul < 1 - 2" in {
    val tryConstruct = Try(Doboz(0,0,"Piros",Vector()))
    assert(tryConstruct.isFailure)
    val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
    assert(tryRecover.isSuccess)
  }

  it should "throw an IllegalArgumentException if tartalom kivul sum > belul" in {
    val tryConstruct = Try(Doboz(10,8,"Piros",Vector(Doboz(5,4,"Kék",Vector()),Doboz(5,4,"Kék",Vector()))))
    assert(tryConstruct.isFailure)
    val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
    assert(tryRecover.isSuccess)
  }
  it should "throw an IllegalArgumentException if tartalom kivul sum > belul  - 2" in {
    val tryConstruct = Try(Doboz(10,8,"Piros",Vector(Doboz(5,4,"Kék",Vector()),Doboz(5,4,"Kék",Vector()))))
    assert(tryConstruct.isFailure)
    val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
    assert(tryRecover.isSuccess)
  }

  it should "set belso meret to 0 if belso < 0" in {
    val doboz = Doboz(10,-1,"Kék", Vector())
    assert(doboz.belul == 0)
  }
  it should "set belso meret to 0 if belso < 0 - 2" in {
    val doboz = Doboz(10,-1,"Kék", Vector())
    assert(doboz.belul == 0)
  }

  it should "set belso meret to kulso-1 if it's too large" in {
    val doboz = Doboz(10,20,"Kék", Vector())
    assert(doboz==Doboz(10,9,"Kék",Vector()))
  }
  it should "set belso meret to kulso-1 if it's too large - 2" in {
    val doboz = Doboz(10,20,"Kék", Vector())
    assert(doboz==Doboz(10,9,"Kék",Vector()))
  }

  it should "check the tartalom against the actual doboz belso meret" in {
    val tryConstruct = Try(Doboz(1,2,"Piros",Vector(Doboz(2,1,"Kék",Vector()))))
    assert(tryConstruct.isFailure)
    val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
    assert(tryRecover.isSuccess)
  }
  it should "check the tartalom against the actual doboz belso meret - 2" in {
    val tryConstruct = Try(Doboz(1,2,"Piros",Vector(Doboz(2,1,"Kék",Vector()))))
    assert(tryConstruct.isFailure)
    val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
    assert(tryRecover.isSuccess)
  }
  it should "check the tartalom against the actual doboz belso meret - 3" in {
    val tryConstruct = Try(Doboz(1,2,"Piros",Vector(Doboz(2,1,"Kék",Vector()))))
    assert(tryConstruct.isFailure)
    val tryRecover = tryConstruct.recover({case e : IllegalArgumentException => ()})
    assert(tryRecover.isSuccess)
  }

  "Doboz kimaradoHely" should "compute one-level kimarado hely" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    assert(doboz.kimaradoHely == 3)
    assert(doboz1.kimaradoHely == 1)
    assert(doboz2.kimaradoHely == 1)
  }
  it should "compute one-level kimarado hely - 2" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    assert(doboz.kimaradoHely == 3)
    assert(doboz1.kimaradoHely == 1)
    assert(doboz2.kimaradoHely == 1)
  }

  "Doboz osszesKimaradoHely" should "compute sum of kimarado helysz" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    assert(doboz.osszesKimaradoHely == 5)
    assert(doboz1.osszesKimaradoHely == 1)
    assert(doboz2.osszesKimaradoHely == 1)
  }
  it should "compute sum of kimarado helysz - 2" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    assert(doboz.osszesKimaradoHely == 5)
    assert(doboz1.osszesKimaradoHely == 1)
    assert(doboz2.osszesKimaradoHely == 1)
  }
  it should "compute sum of kimarado helysz - 3" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    assert(doboz.osszesKimaradoHely == 5)
    assert(doboz1.osszesKimaradoHely == 1)
    assert(doboz2.osszesKimaradoHely == 1)
  }
  it should "compute sum of kimarado helysz - 4" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    assert(doboz.osszesKimaradoHely == 5)
    assert(doboz1.osszesKimaradoHely == 1)
    assert(doboz2.osszesKimaradoHely == 1)
  }

  "Doboz foreach" should "call the function on each contained dobozsz" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val builder = new StringBuilder
    doboz.foreach(
      x => builder.append(x.szin)
    )
    assert(builder.toString == "KékKékKék")
  }
  it should "call the function on each contained dobozsz - 2" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val builder = new StringBuilder
    doboz.foreach(
      x => builder.append(x.szin)
    )
    assert(builder.toString == "KékKékKék")
  }
  it should "call the function on each contained dobozsz - 3" in {
    val doboz1 = Doboz(2,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 3, tartalom=Vector(doboz1))
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val builder = new StringBuilder
    doboz.foreach(
      x => builder.append(x.szin)
    )
    assert(builder.toString == "KékKékKékKék")
  }

  "Doboz filter" should "keep the dobozsz satisfying the predicate" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val filteredDoboz = doboz.filter( _.kivul >= 3 )
    assert(filteredDoboz contains Doboz(10,8,"Kék",Vector(doboz1)))
  }
  it should "keep the dobozsz satisfying the predicate - 2" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val filteredDoboz = doboz.filter( _.kivul >= 3 )
    assert(filteredDoboz contains Doboz(10,8,"Kék",Vector(doboz1)))
  }
  it should "keep the dobozsz satisfying the predicate - 3" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val filteredDoboz = doboz.filter( _.kivul >= 3 )
    assert(filteredDoboz contains Doboz(10,8,"Kék",Vector(doboz1)))
  }

  it should "return None if the outermost Doboz does not satisfy p" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val filteredDoboz = doboz.filter( _.kivul >= 30 )
    assert(filteredDoboz.isEmpty )
  }
  it should "return None if the outermost Doboz does not satisfy p - 2" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val filteredDoboz = doboz.filter( _.kivul >= 30 )
    assert(filteredDoboz.isEmpty )
  }
  it should "return None if the outermost Doboz does not satisfy p - 3" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2)
    val doboz = Doboz(10,8,"Kék",Vector(doboz1,doboz2))
    val filteredDoboz = doboz.filter( _.kivul >= 30 )
    assert(filteredDoboz.isEmpty )
  }

  "paintThemAll" should "set the color of all the inner dobozsz to the parameter" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val festettDoboz = doboz.paintThemAll("Lila")
    assert(festettDoboz == Doboz(10,8,"Lila", Vector(Doboz(3,1,"Lila",Vector()),Doboz(2,1,"Lila",Vector()))))
  }
  it should "set the color of all the inner dobozsz to the parameter - 2" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val festettDoboz = doboz1.paintThemAll("Lila")
    assert(festettDoboz == doboz1.copy(szin="Lila"))
  }
  it should "set the color of all the inner dobozsz to the parameter - 3" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val festettDoboz = doboz.paintThemAll("Lila")
    assert(festettDoboz == Doboz(10,8,"Lila", Vector(Doboz(3,1,"Lila",Vector()),Doboz(2,1,"Lila",Vector()))))
  }

  "colors" should "collect the colors of all the szubdobozsz" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val dobozColors = doboz.colors
    assert(dobozColors == Set("Kék","Fehér","Piros"))
  }
  it should "collect the colors of all the szubdobozsz - 2" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val dobozColors = doboz.colors
    assert(dobozColors == Set("Kék","Fehér","Piros"))
  }
  it should "collect the colors of all the szubdobozsz - 3" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val dobozColors = doboz.colors
    assert(dobozColors == Set("Kék","Fehér","Piros"))
  }
  it should "collect the colors of all the szubdobozsz - 4" in {
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val dobozColors = doboz.colors
    assert(dobozColors == Set("Kék","Fehér","Piros"))
  }

  "save" should "print out the contents of the doboz according to the specification" in {
    val writer = new PrintWriter("dobozTest.txt")
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val trySave = DobozSaver(writer).save(doboz)
    writer.close()
    assert(trySave.isSuccess)
    val lines = Source.fromFile("dobozTest.txt")
    val lineVector = lines.getLines().toVector
    lines.close()
    assert(lineVector == Vector("[ 10;8;Piros", "[ 3;1;Kék", "]", "[ 2;1;Fehér", "]","]"))
  }
  it should "print out the contents of the doboz according to the specification - 2" in {
    val writer = new PrintWriter("dobozTest.txt")
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val trySave = DobozSaver(writer).save(doboz)
    writer.close()
    assert(trySave.isSuccess)
    val lines = Source.fromFile("dobozTest.txt")
    val lineVector = lines.getLines().toVector
    lines.close()
    assert(lineVector == Vector("[ 10;8;Piros", "[ 3;1;Kék", "]", "[ 2;1;Fehér", "]","]"))
  }
  it should "print out the contents of the doboz according to the specification - 3" in {
    val writer = new PrintWriter("dobozTest.txt")
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val trySave = DobozSaver(writer).save(doboz)
    writer.close()
    assert(trySave.isSuccess)
    val lines = Source.fromFile("dobozTest.txt")
    val lineVector = lines.getLines().toVector
    lines.close()
    assert(lineVector == Vector("[ 10;8;Piros", "[ 3;1;Kék", "]", "[ 2;1;Fehér", "]","]"))
  }
  it should "print out the contents of the doboz according to the specification - 4" in {
    val writer = new PrintWriter("dobozTest.txt")
    val doboz1 = Doboz(3,1,"Kék",Vector())
    val doboz2 = doboz1.copy(kivul = 2,szin="Fehér")
    val doboz = Doboz(10,8,"Piros",Vector(doboz1,doboz2))
    val trySave = DobozSaver(writer).save(doboz)
    writer.close()
    assert(trySave.isSuccess)
    val lines = Source.fromFile("dobozTest.txt")
    val lineVector = lines.getLines().toVector
    lines.close()
    assert(lineVector == Vector("[ 10;8;Piros", "[ 3;1;Kék", "]", "[ 2;1;Fehér", "]","]"))
  }


}
